@@ -4,9 +4,10 @@ from io import UnsupportedOperation
 import numpy as np
 import torch
 
-from sidechainnet.utils.sequence import VOCAB
+from sidechainnet.utils.sequence import ONE_TO_THREE_LETTER_MAP, VOCAB
 from sidechainnet.structure.build_info import SC_BUILD_INFO, BB_BUILD_INFO, NUM_COORDS_PER_RES, SC_ANGLES_START_POS, NUM_ANGLES
 from sidechainnet.structure.structure import nerf
+from sidechainnet.structure.HydrogenBuilder import HydrogenBuilder, NUM_COORDS_PER_RES_W_HYDROGENS
 
 
 class StructureBuilder(object):
@@ -65,7 +66,7 @@ class StructureBuilder(object):
             raise ValueError("You must provide exactly one of either coordinates (crd) "
                              "or angles (ang).")
 
-        self.seq_as_str = seq if type(seq) == str else self._convert_seq_to_str(seq)
+        self.seq_as_str = seq if type(seq) == str else _convert_seq_to_str(seq)
         self.seq_as_ints = np.asarray([VOCAB._char2int[s] for s in self.seq_as_str])
         self.device = device
 
@@ -100,27 +101,9 @@ class StructureBuilder(object):
         self.next_bb = None
         self.pdb_creator = None
         self.nerf_method = nerf_method
-
-    @staticmethod
-    def _convert_seq_to_str(seq):
-        """Assuming seq is an int list or int tensor, returns its str representation."""
-        seq_as_str = ""
-        if isinstance(seq, torch.Tensor):
-            seq = seq.numpy()
-        seq = seq.flatten()
-        if len(seq.shape) == 1:
-            # The seq is represented as an integer sequence
-            if len(seq.shape) == 1:
-                seq_as_str = VOCAB.ints2str(seq)
-            elif len(seq.shape) == 2:
-                if seq.shape[0] != 1:
-                    raise ValueError(f"Seq shape {seq.shape} is not supported.")
-                else:
-                    seq_as_str = VOCAB.ints2str(seq[0])
-        else:
-            raise UnsupportedOperation(f"Seq shape {seq.shape} is not supported.")
-
-        return seq_as_str
+        self.has_hydrogens = False
+        self.atoms_per_res = NUM_COORDS_PER_RES
+        self.terminal_atoms = None
 
     def __len__(self):
         """Return length of the protein sequence.
@@ -198,10 +181,25 @@ class StructureBuilder(object):
         if not self.pdb_creator:
             from sidechainnet.structure.PdbBuilder import PdbBuilder
             if self.data_type == 'numpy':
-                self.pdb_creator = PdbBuilder(self.seq_as_str, self.coords)
+                self.pdb_creator = PdbBuilder(self.seq_as_str, self.coords,
+                                              self.atoms_per_res,
+                                              terminal_atoms=self.terminal_atoms)
             else:
                 self.pdb_creator = PdbBuilder(self.seq_as_str,
-                                              self.coords.detach().numpy())
+                                              self.coords.detach().numpy(),
+                                              self.atoms_per_res,
+                                              terminal_atoms=self.terminal_atoms)
+
+    def add_hydrogens(self):
+        """Add Hydrogen atom coordinates to coordinate representation (re-apply PADs)."""
+        if self.coords is None or not len(self.coords):
+            raise ValueError("Cannot add hydrogens to a structure whose heavy atoms have"
+                             " not yet been built.")
+        self.hb = HydrogenBuilder(self.seq_as_str, self.coords)
+        self.coords = self.hb.build_hydrogens()
+        self.has_hydrogens = True
+        self.atoms_per_res = NUM_COORDS_PER_RES_W_HYDROGENS
+        self.terminal_atoms = self.hb.terminal_atoms
 
     def to_pdb(self, path, title="pred"):
         """Save protein structure as a PDB file to given path.
@@ -213,6 +211,15 @@ class StructureBuilder(object):
         self._initialize_coordinates_and_PdbCreator()
         self.pdb_creator.save_pdb(path, title)
 
+    def to_pdbstr(self, title="pred"):
+        """Return protein structure as a PDB string.
+
+        Args:
+            title (str, optional): Title of structure for PDB file. Defaults to "pred".
+        """
+        self._initialize_coordinates_and_PdbCreator()
+        return self.pdb_creator.get_pdb_string(title)
+
     def to_gltf(self, path, title="pred"):
         """Save protein structure as a GLTF (3D-object) file to given path.
 
@@ -237,10 +244,9 @@ class StructureBuilder(object):
         import py3Dmol
         if not style:
             style = {'cartoon': {'color': 'spectrum'}, 'stick': {'radius': .15}}
-        self._initialize_coordinates_and_PdbCreator()
 
         view = py3Dmol.view(**kwargs)
-        view.addModel(self.pdb_creator.get_pdb_string(), 'pdb')
+        view.addModel(self.to_pdbstr(), 'pdb', {'keepH': True})
         if style:
             view.setStyle(style)
         view.zoomTo()
@@ -275,7 +281,7 @@ class ResidueBuilder(object):
                 "standard".
         """
         if (not isinstance(name, np.int64) and not isinstance(name, np.int32) and
-                not isinstance(name, torch.Tensor)):
+                not isinstance(name, int) and not isinstance(name, torch.Tensor)):
             raise ValueError("Expected integer AA code." + str(name.shape) +
                              str(type(name)))
         if isinstance(angles, np.ndarray):
@@ -291,7 +297,12 @@ class ResidueBuilder(object):
         self.bb = []
         self.sc = []
         self.coords = []
-        self.coordinate_padding = torch.zeros(3)
+        self.coordinate_padding = torch.zeros(3, requires_grad=True)
+
+    @property
+    def AA(self):
+        """Return the one-letter amino acid code (str) for this residue."""
+        return VOCAB.int2char(int(self.name))
 
     def build(self):
         """Construct and return atomic coordinates for this protein."""
@@ -355,12 +366,12 @@ class ResidueBuilder(object):
 
         Placed in an arbitrary plane (z = .001).
         """
-        n = torch.tensor([0, 0, 0.001], device=self.device)
+        n = torch.tensor([0, 0, 0.001], device=self.device, requires_grad=True)
         ca = n + torch.tensor([BB_BUILD_INFO["BONDLENS"]["n-ca"], 0, 0],
-                              device=self.device)
+                              device=self.device, requires_grad=True)
         cx = torch.cos(np.pi - self.ang[3]) * BB_BUILD_INFO["BONDLENS"]["ca-c"]
         cy = torch.sin(np.pi - self.ang[3]) * BB_BUILD_INFO["BONDLENS"]['ca-c']
-        c = ca + torch.tensor([cx, cy, 0], device=self.device, dtype=torch.float32)
+        c = ca + torch.tensor([cx, cy, 0], device=self.device, dtype=torch.float32, requires_grad=True)
         o = nerf(
             n,
             ca,
@@ -380,6 +391,7 @@ class ResidueBuilder(object):
         neighboring (previous or next) residue.
         """
         assert len(self.bb) > 0, "Backbone must be built first."
+        self.atom_names = ["N", "CA", "C", "O"]
         self.pts = {"N": self.bb[0], "CA": self.bb[1], "C": self.bb[2]}
         if self.next_res:
             self.pts["N+"] = self.next_res.bb[0]
@@ -414,6 +426,7 @@ class ResidueBuilder(object):
             self.pts[atom_names[-1]] = new_pt
             self.sc.append(new_pt)
             last_torsion = torsion
+            self.atom_names.append(atom_names[-1])
 
         return self.sc
 
@@ -422,9 +435,19 @@ class ResidueBuilder(object):
                                            len(self.sc)) * [self.coordinate_padding]
         return self.coords
 
+    def to_prody(self, res):
+        import prody as pr
+        ag = pr.AtomGroup()
+        ag.setCoords(torch.stack(self.bb + self.sc).numpy())
+        ag.setNames(self.atom_names)
+        ag.setResnames([ONE_TO_THREE_LETTER_MAP[VOCAB._int2char[self.name]]] *
+                       len(self.atom_names))
+        ag.setResnums([res.getResnum()] * len(self.atom_names))
+        return pr.Residue(ag, [0] * len(self.atom_names), None)
+
     def __repr__(self):
         """Return a string describing the name of the residue used for this object."""
-        return f"ResidueBuilder({VOCAB.int2char(int(self.name))})"
+        return f"ResidueBuilder({self.AA})"
 
 
 def _get_residue_build_iter(res, build_dictionary):
@@ -463,6 +486,27 @@ def _get_residue_build_iter(res, build_dictionary):
             [t.split("-") for t in r["torsion-names"]]))
 
 
+def _convert_seq_to_str(seq):
+    """Assuming seq is an int list or int tensor, returns its str representation."""
+    seq_as_str = ""
+    if isinstance(seq, torch.Tensor):
+        seq = seq.numpy()
+    seq = seq.flatten()
+    if len(seq.shape) == 1:
+        # The seq is represented as an integer sequence
+        if len(seq.shape) == 1:
+            seq_as_str = VOCAB.ints2str(seq)
+        elif len(seq.shape) == 2:
+            if seq.shape[0] != 1:
+                raise ValueError(f"Seq shape {seq.shape} is not supported.")
+            else:
+                seq_as_str = VOCAB.ints2str(seq[0])
+    else:
+        raise UnsupportedOperation(f"Seq shape {seq.shape} is not supported.")
+
+    return seq_as_str
+
+
 if __name__ == '__main__':
     import pickle
 
